/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.bc.ftp.service;

import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;

import org.ow2.petals.bc.ftp.FTPConstants;
import org.ow2.petals.bc.ftp.FTPUtil;
import org.ow2.petals.bc.ftp.MissingElementException;
import org.ow2.petals.bc.ftp.connection.FTPConnection;
import org.ow2.petals.bc.ftp.connection.FTPConnectionInfo;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * The FTP JBI listener. Writes a file to FTP server on incoming JBI message.
 * 
 * @author alouis
 */
public class FTPService {

    private final Logger logger;

    public FTPService(final Logger logger) {
        super();
        this.logger = logger;
    }

    /**
     * Put the source on the ftp server with the given filename a Fault is
     * returned on IO problems TODO the content to send is a String, which is
     * not optimal
     * 
     * @param exchange
     * @param fileName
     * @param connectionInfo
     * @param inMessage
     * @throws JBIException
     *             bad MEP
     * @throws MissingElementException
     */
    public void processPut(final Exchange exchange, final FTPConnectionInfo connectionInfo,
            final Document inMessage) throws JBIException, MissingElementException {
        if (exchange == null) {
            throw new NullPointerException("exchange parameter is null");
        }

        if (this.logger.isLoggable(Level.INFO)) {
            this.logger.info("JBI exchange received - Process PUT operation");
        }

        // Checks the MEP of the incoming message
        if (!exchange.isInOnlyPattern()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Exchange pattern for the PUT operation must be ");
            errorSB.append(Message.MEPConstants.IN_ONLY_PATTERN.value());
            throw new JBIException(errorSB.toString());
        }
        // Check the root node
        final Node root = inMessage.getFirstChild();
        if (root == null || !FTPConstants.PUT_OPERATION.equalsIgnoreCase(root.getLocalName())) {
            throw new MissingElementException(FTPConstants.PUT_OPERATION);
        }

        final Node fileNameNode = XMLUtil.findChild(root, FTPConstants.FTP_FILENAME, null, false);
        if (fileNameNode == null
                || !FTPConstants.FTP_FILENAME.equalsIgnoreCase(fileNameNode.getLocalName())) {
            throw new MissingElementException(FTPConstants.FTP_FILENAME);
        }
        final Node contentNode = XMLUtil.findChild(root, FTPConstants.BODY, null, false);
        if (contentNode == null) {
            throw new MissingElementException(FTPConstants.BODY);
        }
        try {
            final Node rootBody = contentNode.getFirstChild();
            String body;
            if (rootBody != null) {
                body = XMLUtil.createStringFromDOMNode(rootBody);
            } else {
                body = contentNode.getTextContent();
            }
            final FTPConnection ftpConnection = new FTPConnection(connectionInfo, this.logger);
            ftpConnection.putString(fileNameNode.getTextContent(), body);
        } catch (final IOException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : IOException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        } catch (final TransformerException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : TransformerException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        } catch (DOMException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : DOMException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        } catch (InterruptedException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : InterruptedException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("PUT operation ended");
        }
    }

    /**
     * Put the body element of the XML message on the ftp server. a Fault is
     * returned on IO problems TODO the content to send is a String, which is
     * not optimal
     * 
     * @param exchange
     * @param fileName
     * @param connectionInfo
     * @param inMessage
     * @throws JBIException
     *             bad MEP
     */
    public void processStandardPut(final Exchange exchange, final String fileName,
            final FTPConnectionInfo connectionInfo, final Document inMessage) throws JBIException {
        if (exchange == null) {
            throw new NullPointerException("exchange parameter is null");
        }
        if (this.logger.isLoggable(Level.INFO)) {
            this.logger
                    .info("JBI exchange received - Process standard PUT operation (the whole XML message put on the FTP server)");
        }
        // Checks the MEP of the incoming message
        if (!exchange.isInOnlyPattern()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Exchange pattern for the PUT operation must be ");
            errorSB.append(Message.MEPConstants.IN_ONLY_PATTERN.value());
            throw new JBIException(errorSB.toString());
        }

        try {
            final String content = XMLUtil.createStringFromDOMNode(inMessage);
            final FTPConnection ftpConnection = new FTPConnection(connectionInfo, this.logger);
            ftpConnection.putString(fileName, content);

        } catch (final IOException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : IOException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        } catch (final TransformerException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : TransformerException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        } catch (InterruptedException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : InterruptedException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("PUT operation ended");
        }
    }

    /**
     * Put attachments on the ftp server TODO an FTP connection is done for each
     * file...
     * 
     * @param exchange
     * @param connectionInfo
     * @throws JBIException
     */
    public void processMPut(final Exchange exchange, final FTPConnectionInfo connectionInfo)
            throws JBIException {
        if (this.logger.isLoggable(Level.INFO)) {
            this.logger.info("JBI exchange received - Process MPUT operation");
        }
        // Checks the MEP of the incoming message
        if (!exchange.isInOnlyPattern()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Exchange pattern for the MPUT operation must be ");
            errorSB.append(Message.MEPConstants.IN_ONLY_PATTERN.value());
            throw new JBIException(errorSB.toString());
        }
        
        // Check if the message has attachements
        if (exchange.getInMessageAttachments().isEmpty()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("One or more attachments expected.");
            throw new JBIException(errorSB.toString());
        }
        
        try {
            final FTPConnection ftpConnection = new FTPConnection(connectionInfo, this.logger);

            // rebuild a hashmap containing files and their names
            final Map<String, DataHandler> attachments = new HashMap<String, DataHandler>();
            for (final Object fileName : exchange.getInMessage().getAttachmentNames()) {
                attachments.put((String) fileName, exchange.getInMessage().getAttachment(
                        (String) fileName));
            }
            ftpConnection.mPut(attachments);
        } catch (IOException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : IOException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        } catch (InterruptedException e) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on FTP server [");
            errorSB.append(connectionInfo.getServer());
            errorSB.append("] : InterruptedException : ");
            errorSB.append(e.getMessage());
            throw new JBIException(errorSB.toString());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("MPUT operation ended");
        }
    }

    /**
     * FTP DIR operation. <br>
     * a Fault is returned for bad request or IO problems
     * 
     * @param exchange
     *            the message exchange, MUST be non null
     * @throws JBIException
     *             bad MEP
     */

    public void processDir(final Exchange exchange, final FTPConnectionInfo connectionInfo)
            throws JBIException {
        if (this.logger.isLoggable(Level.INFO)) {
            this.logger.info("JBI exchange received - Process DIR operation");
        }
        if (exchange.isInOnlyPattern() || exchange.isRobustInOnlyPattern()) {
            throw new JBIException("Operation DIR " + "' must use the pattern '"
                    + Message.MEPConstants.IN_OUT_PATTERN.value() + "'");
        }
        List<String> fileNames;
        try {
            // do the FTP DIR operation
            final FTPConnection ftpConnection = new FTPConnection(connectionInfo, this.logger);
            fileNames = ftpConnection.list();

            // generate filename list
            final String result = FTPUtil.generateFileNameList(fileNames, "dir");
            exchange.getOutMessage().setContent(UtilFactory.getSourceUtil().createSource(result));
        } catch (Exception e) {
            FTPUtil.setIOFaultOnExchange(exchange, "Can not list files on FTP server ["
                    + connectionInfo.getServer() + "]. Cause : " + e.getMessage());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("DIR operation ended");
        }
    }

    /**
     * FTP MGET operation <br>
     * All files are getted as attachments
     * 
     * @param exchange
     * @param connectionInfo
     * @param inMessage
     * @throws Exception
     */
    public void processMGet(final Exchange exchange, final FTPConnectionInfo connectionInfo,
            final Document inMessage) throws Exception {
        if (this.logger.isLoggable(Level.INFO)) {
            this.logger.info("JBI exchange received - Process MGET operation");
        }
        if (exchange.isInOnlyPattern() || exchange.isRobustInOnlyPattern()) {
            throw new JBIException("Operation MGET " + "' must use the pattern '"
                    + Message.MEPConstants.IN_OUT_PATTERN.value() + "'");
        }
        // retrieve filenames to get
        final NodeList list = inMessage.getFirstChild().getChildNodes();
        final List<String> filenames = new LinkedList<String>();
        for (int index = 0; index < list.getLength(); index++) {
            if (FTPConstants.FTP_FILENAME.equalsIgnoreCase(list.item(index).getLocalName())) {
                filenames.add(list.item(index).getTextContent());
            }
        }
        // do the FTP MGET operation
        final FTPConnection ftpConnection = new FTPConnection(connectionInfo, this.logger);
        final Map<String, DataHandler> files = ftpConnection.mGet(filenames);

        // add attachments on the OUT message
        for (final String filename : files.keySet()) {
            exchange.getOutMessage().addAttachment(filename, files.get(filename));
        }
        final List<String> retrievedFilenames = new LinkedList<String>(files.keySet());
        final String response = FTPUtil.generateFileNameList(retrievedFilenames, "mget");
        exchange.getOutMessage().setContent(UtilFactory.getSourceUtil().createSource(response));
        if (connectionInfo.getDeleteProcessedFile()) {
            ftpConnection.mDel(retrievedFilenames);
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("MGET operation ended");
        }
    }

    /**
     * 
     * FTP GET operation <br>
     * get the file as a source message a Fault is returned for bad request or
     * IO problems
     * 
     * @param exchange
     *            the message exchange, MUST be non null
     * @param connectionInfo
     * @param inMessage
     * @throws JBIException
     *             bad MEP
     */
    public void processGet(final Exchange exchange, final FTPConnectionInfo connectionInfo,
            final Document inMessage) throws JBIException {
        if (this.logger.isLoggable(Level.INFO)) {
            this.logger.info("JBI exchange received - Process GET operation");
        }
        // Checks the MEP of the incoming message
        if (exchange.isInOnlyPattern() || exchange.isRobustInOnlyPattern()) {
            throw new JBIException("Operation GET " + "' must use the pattern '"
                    + Message.MEPConstants.IN_OUT_PATTERN.value() + "'");
        }
        // retrieve the filename to get
        final Node filePatternNode = XMLUtil.findChild(inMessage.getFirstChild(),
                FTPConstants.FTP_FILENAME, null, false);
        String filePattern = null;
        if (filePatternNode != null) {
            filePattern = filePatternNode.getTextContent();
        } else {
            FTPUtil.setMissingElementFaultOnExchange(exchange, FTPConstants.FTP_FILENAME);
        }

        if (filePattern != null) {
            try {
                // do the FTP GET operation
                final FTPConnection ftpConnection = new FTPConnection(connectionInfo, this.logger);
                final Source stream = ftpConnection.get(filePattern);
                exchange.setOutMessageContent(stream);
                if (connectionInfo.getDeleteProcessedFile()) {
                    ftpConnection.del(filePattern);
                }
            } catch (final Exception e) {
                FTPUtil.setIOFaultOnExchange(exchange, "Can not get file [" + filePattern
                        + "] on FTP server [" + connectionInfo.getServer() + "]. Cause : "
                        + e.getMessage());
            }
        } else {
            FTPUtil.setMissingElementFaultOnExchange(exchange, FTPConstants.FTP_FILENAME);
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("GET operation ended");
        }
    }

    /***
     * 
     * @param exchange
     * @param connectionInfo
     * @param inMessage
     * @throws Exception
     */
    public void processDel(final Exchange exchange, final FTPConnectionInfo connectionInfo,
            final Document inMessage) throws Exception {
        if (this.logger.isLoggable(Level.INFO)) {
            this.logger.info("JBI exchange received - Process DEL operation");
        }
        // Checks the MEP of the incoming message
        if (!exchange.isInOnlyPattern()) {
            throw new JBIException("Operation DEL " + "' must use the pattern '"
                    + Message.MEPConstants.IN_ONLY_PATTERN.value() + "'");
        }
        // retrieve the filename to get
        final Node filePatternNode = XMLUtil.findChild(inMessage.getFirstChild(),
                FTPConstants.FTP_FILENAME, null, false);
        String filePattern = null;
        if (filePatternNode != null) {
            filePattern = filePatternNode.getTextContent();
        } else {
            throw new MissingElementException(FTPConstants.FTP_FILENAME);
        }

        if (filePattern != null) {
            try {
                // do the FTP DEL operation
                final FTPConnection ftpConnection = new FTPConnection(connectionInfo, this.logger);
                ftpConnection.del(filePattern);
            } catch (final IOException e) {
                final StringBuilder errorSB = new StringBuilder();
                errorSB.append("Can not delete file on FTP server [");
                errorSB.append(connectionInfo.getServer());
                errorSB.append("] : IOException : ");
                errorSB.append(e.getMessage());
                throw new JBIException(errorSB.toString());
            }
            if (this.logger.isLoggable(Level.FINEST)) {
                this.logger.finest("DEL operation ended");
            }
        }
    }
}
